@@ -1,12 +1,11 @@
 /*
  * This file is part of the MQT QMAP library which is released under the MIT license.
- * See file README.md or go to https://iic.jku.at/eda/research/ibm_qx_mapping/ for more information.
+ * See file README.md or go to https://www.cda.cit.tum.de/research/ibm_qx_mapping/ for more information.
  */
 
 #ifndef EXACT_MAPPER_hpp
 #define EXACT_MAPPER_hpp
 
-#include "Encodings.hpp"
 #include "Mapper.hpp"
 
 #include <algorithm>
@@ -16,10 +15,7 @@
 #include <functional>
 #include <set>
 #include <unordered_set>
-#include <z3++.h>
 
-using namespace z3;
-using matrix      = std::vector<expr_vector>;
 using Swaps       = std::vector<std::pair<unsigned short, unsigned short>>;
 using QubitChoice = std::set<unsigned short>;
 
