@@ -1,13 +1,17 @@
 /*
  * This file is part of the MQT QMAP library which is released under the MIT license.
- * See file README.md or go to https://iic.jku.at/eda/research/ibm_qx_mapping/ for more information.
+ * See file README.md or go to https://www.cda.cit.tum.de/research/ibm_qx_mapping/ for more information.
  */
 
 #include "Architecture.hpp"
 
+#include "utils.hpp"
+
+#include <utility>
+
 void Architecture::loadCouplingMap(AvailableArchitecture architecture) {
     std::stringstream ss{getCouplingMapSpecification(architecture)};
-    architectureName = toString(architecture);
+    name = toString(architecture);
     loadCouplingMap(ss);
 }
 
@@ -16,10 +20,10 @@ void Architecture::loadCouplingMap(std::istream& is) {
 }
 
 void Architecture::loadCouplingMap(const std::string& filename) {
-    size_t slash     = filename.find_last_of('/');
-    size_t dot       = filename.find_last_of('.');
-    architectureName = filename.substr(slash + 1, dot - slash - 1);
-    auto ifs         = std::ifstream(filename);
+    size_t slash = filename.find_last_of('/');
+    size_t dot   = filename.find_last_of('.');
+    name         = filename.substr(slash + 1, dot - slash - 1);
+    auto ifs     = std::ifstream(filename);
     if (ifs.good())
         this->loadCouplingMap(ifs);
     else
@@ -28,7 +32,7 @@ void Architecture::loadCouplingMap(const std::string& filename) {
 
 void Architecture::loadCouplingMap(std::istream&& is) {
     couplingMap.clear();
-    calibrationData.clear();
+    properties.clear();
     std::string line;
 
     std::regex  r_nqubits = std::regex("([0-9]+)");
@@ -70,43 +74,105 @@ void Architecture::loadCouplingMap(std::istream&& is) {
 void Architecture::loadCouplingMap(unsigned short nQ, const CouplingMap& cm) {
     nqubits     = nQ;
     couplingMap = cm;
-    calibrationData.clear();
-    architectureName = "generic_" + std::to_string(nQ);
+    properties.clear();
+    name = "generic_" + std::to_string(nQ);
     createDistanceTable();
 }
 
-void Architecture::loadCalibrationData(std::istream& is) {
-    loadCalibrationData(std::move(is));
+void Architecture::loadProperties(std::istream& is) {
+    loadProperties(std::move(is));
 }
 
-void Architecture::loadCalibrationData(const std::string& filename) {
-    size_t slash    = filename.find_last_of('/');
-    size_t dot      = filename.find_last_of('.');
-    calibrationName = filename.substr(slash + 1, dot - slash - 1);
-    auto ifs        = std::ifstream(filename);
+void Architecture::loadProperties(const std::string& filename) {
+    size_t slash = filename.find_last_of('/');
+    size_t dot   = filename.find_last_of('.');
+    properties.setName(filename.substr(slash + 1, dot - slash - 1));
+    if (!isArchitectureAvailable())
+        name = properties.getName();
+    auto ifs = std::ifstream(filename);
     if (ifs.good())
-        this->loadCalibrationData(ifs);
+        this->loadProperties(ifs);
     else
-        throw QMAPException("Error opening calibration data file.");
+        throw QMAPException("Error opening properties file.");
 }
 
-void Architecture::loadCalibrationData([[maybe_unused]] std::istream&& is) {
-    calibrationData.clear();
-    calibrationData.reserve(nqubits);
-    // TODO: Read in IBM csv Format into calibrationData
-    // Qubit,T1 (µs),T2 (µs),Frequency (GHz),Readout error,Single-qubit U2 error rate,CNOT error rate,Date
-    // Q0,43.113390027720456,80.43170304899623,5.2540548421267825,3.0000000000000027e-2,6.471748738850555e-4,"cx0_1: 1.286e-2",Fri Feb 28 2020 09:27:24 GMT+0100 (CET)
-    // Q1,57.036278779745814,44.25327200058967,5.04877069952246,5.499999999999994e-2,6.227987234977298e-4,"cx1_0: 1.286e-2, cx1_2: 1.117e-2, cx1_3: 1.006e-2",Fri Feb 28 2020 09:27:24 GMT+0100 (CET)
-    // Q2,81.96027845904459,87.47080521889637,5.230569177436213,2.1666666666666723e-2,5.16080288896023e-4,"cx2_1: 1.117e-2",Fri Feb 28 2020 09:27:24 GMT+0100 (CET)
-    // Q3,67.95013744985455,86.03784184638282,5.200948312700143,1.8333333333333313e-2,3.530411094454146e-4,"cx3_1: 1.006e-2, cx3_4: 1.670e-2",Fri Feb 28 2020 09:27:24 GMT+0100 (CET)
-    // Q4,62.156360190670306,3.731765939150933,5.065777885569538,1.5000000000000013e-2,1.106360359568795e-3,"cx4_3: 1.670e-2",Fri Feb 28 2020 09:27:24 GMT+0100 (CET)
+void Architecture::loadProperties(std::istream&& is) {
+    static const auto SingleQubitGates = {"id", "u1", "u2", "u3", "rz", "sx", "x"};
+
+    properties.clear();
+
+    double averageCNOTFidelity = 0.0;
+    int    numCNOTFidelities   = 0;
+
+    std::string line;
+    std::string word;
+    std::regex  regexDoubleFidelity =
+            std::regex(R"(((\d+).?(\d+):\W*?(-?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+-]?\d+)?)))");
+    std::smatch sMatch;
+    std::getline(is, line); //skip first line
+    // load edges
+    int qubitNumber = 0;
+    while (std::getline(is, line)) {
+        std::stringstream        ss(line);
+        std::vector<std::string> data{};
+        parse_line(line, ',', {'\"'}, {'\\'}, data);
+        properties.t1Time.set(qubitNumber, std::stod(data.at(1)));
+        properties.t2Time.set(qubitNumber, std::stod(data.at(2)));
+        properties.qubitFrequency.set(qubitNumber, std::stod(data.at(3)));
+        properties.readoutErrorRate.set(qubitNumber, std::stod(data.at(4)));
+        // csv file reports average single qubit fidelities
+        for (const auto& operation: SingleQubitGates) {
+            properties.setSingleQubitErrorRate(qubitNumber, operation, std::stod(data.at(5)));
+        }
+        // only try to parse CNOT fidelities if there are any
+        if (data.size() >= 7U) {
+            std::string s = data[6];
+            while (std::regex_search(s, sMatch, regexDoubleFidelity)) {
+                auto a = static_cast<unsigned short>(std::stoul(sMatch.str(2U)));
+                auto b = static_cast<unsigned short>(std::stoul(sMatch.str(3U)));
+                if (!isArchitectureAvailable()) {
+                    couplingMap.emplace(a, b);
+                }
+                // calc moving average
+                averageCNOTFidelity = averageCNOTFidelity + (std::stod(sMatch.str(4U)) - averageCNOTFidelity) / ++numCNOTFidelities;
+                properties.setTwoQubitErrorRate(a, b, std::stod(sMatch.str(4U)));
+                s = sMatch.suffix().str();
+            }
+        }
+        if (data.size() == 8U) {
+            properties.calibrationDate.set(qubitNumber, data[7]);
+        }
+        ++qubitNumber;
+    }
+
+    if (isArchitectureAvailable()) {
+        for (const auto& edge: couplingMap) {
+            if (!properties.twoQubitErrorRateAvailable(edge.first, edge.second)) {
+                properties.setTwoQubitErrorRate(edge.first, edge.second, averageCNOTFidelity);
+            }
+        }
+    }
+    properties.setNqubits(qubitNumber);
+    if (!isArchitectureAvailable()) {
+        nqubits = static_cast<unsigned short>(qubitNumber);
+        createDistanceTable();
+    }
 
     createFidelityTable();
 }
 
-void Architecture::loadCalibrationData(const std::vector<CalibrationData>& calData) {
-    calibrationData  = calData;
-    architectureName = "generic_" + std::to_string(nqubits);
+void Architecture::loadProperties(const Properties& props) {
+    if (!isArchitectureAvailable()) {
+        for (const auto& [control, targetProps]: props.twoQubitErrorRate.get()) {
+            for (const auto& [target, errorRate]: targetProps.get()) {
+                couplingMap.emplace(control, target);
+            }
+        }
+        nqubits = props.getNqubits();
+        name    = "generic_" + std::to_string(nqubits);
+        createDistanceTable();
+    }
+    properties = props;
     createFidelityTable();
 }
 
@@ -114,9 +180,9 @@ Architecture::Architecture(unsigned short nQ, const CouplingMap& couplingMap) {
     loadCouplingMap(nQ, couplingMap);
 }
 
-Architecture::Architecture(unsigned short nQ, const CouplingMap& couplingMap, const std::vector<CalibrationData>& calibrationData):
+Architecture::Architecture(unsigned short nQ, const CouplingMap& couplingMap, const Properties& props):
     Architecture(nQ, couplingMap) {
-    loadCalibrationData(calibrationData);
+    loadProperties(props);
 }
 
 void Architecture::createDistanceTable() {
@@ -140,16 +206,18 @@ void Architecture::createDistanceTable() {
 
 void Architecture::createFidelityTable() {
     fidelityTable.clear();
-    fidelityTable.resize(nqubits, std::vector<double>(nqubits, 1.0));
+    fidelityTable.resize(nqubits, std::vector<double>(nqubits, 0.0));
 
-    singleQubitFidelities.resize(nqubits, 1.0);
+    singleQubitFidelities.resize(nqubits, 0.0);
 
-    for (const auto& qubit: calibrationData) {
-        for (const auto& entry: qubit.cnotErrorRate) {
-            fidelityTable.at(entry.first.first).at(entry.first.second) -= entry.second;
+    for (const auto& [first, second]: couplingMap) {
+        if (properties.twoQubitErrorRateAvailable(first, second)) {
+            fidelityTable[first][second] = 1.0 - properties.getTwoQubitErrorRate(first, second);
         }
-        singleQubitFidelities.at(qubit.qubit) -= qubit.singleQubitErrorRate;
     }
+
+    for (const auto& [qubit, operationProps]: properties.singleQubitErrorRate.get())
+        singleQubitFidelities[qubit] = 1.0 - properties.getAverageSingleQubitErrorRate(qubit);
 }
 
 unsigned long Architecture::minimumNumberOfSwaps(std::vector<unsigned short>& permutation, long limit) {
@@ -240,6 +308,10 @@ void Architecture::minimumNumberOfSwaps(std::vector<unsigned short>& permutation
         qubits.insert(q);
     }
 
+    if (qubits.size() != permutation.size()) {
+        throw std::runtime_error("Architecture::minimumNumberOfSwaps: permutation contains duplicates");
+    }
+
     // create map for goal permutation
     std::unordered_map<unsigned short, unsigned short> goalPermutation{};
     unsigned short                                     count    = 0;
@@ -457,3 +529,99 @@ void Architecture::findCouplingLimit(unsigned short node, int curSum, const std:
 
     visited[node] = false;
 }
+
+void Architecture::getHighestFidelityCouplingMap(unsigned short subsetSize, CouplingMap& reducedMap) {
+    if (!isArchitectureAvailable() || nqubits == subsetSize || properties.empty()) {
+        reducedMap = couplingMap;
+    } else {
+        double bestFidelity        = 0.0;
+        auto   allConnectedSubsets = getAllConnectedSubsets(subsetSize);
+
+        for (const auto& qubitChoice: allConnectedSubsets) {
+            double      currentFidelity{};
+            CouplingMap map{};
+            getReducedCouplingMap(qubitChoice, map);
+            currentFidelity = getAverageArchitectureFidelity(map, qubitChoice, properties);
+            if (currentFidelity > bestFidelity) {
+                reducedMap   = map;
+                bestFidelity = currentFidelity;
+            }
+        }
+    }
+}
+std::vector<std::set<unsigned short>> Architecture::getAllConnectedSubsets(unsigned short subsetSize) {
+    std::vector<std::set<unsigned short>> result{};
+    if (!isArchitectureAvailable() || nqubits == subsetSize) {
+        result.emplace_back(getQubitSet());
+    } else if (nqubits < subsetSize) {
+        throw QMAPException("Architecture too small!");
+    } else {
+        auto filter = [&](const std::set<unsigned short>& subset) {
+            CouplingMap cm = {};
+            Architecture::getReducedCouplingMap(subset, cm);
+            return isConnected(subset, cm);
+        };
+        for (const auto& subset: subsets(getQubitSet(), subsetSize, filter)) {
+            result.emplace_back(subset);
+        }
+    }
+    return result;
+}
+void Architecture::getReducedCouplingMaps(unsigned short subsetSize, std::vector<CouplingMap>& couplingMaps) {
+    couplingMaps.clear();
+    if (!isArchitectureAvailable()) {
+        couplingMaps.emplace_back(getFullyConnectedMap(subsetSize));
+    } else {
+        for (const auto& qubitChoice: getAllConnectedSubsets(subsetSize)) {
+            couplingMaps.emplace_back();
+            getReducedCouplingMap(qubitChoice, couplingMaps.back());
+        }
+    }
+}
+void Architecture::getReducedCouplingMap(const std::set<unsigned short>& qubitChoice, CouplingMap& reducedMap) {
+    reducedMap.clear();
+    if (!isArchitectureAvailable()) {
+        reducedMap = getFullyConnectedMap(qubitChoice.size());
+    } else {
+        for (const auto& [q0, q1]: couplingMap) {
+            if (qubitChoice.find(q0) != qubitChoice.end() && qubitChoice.find(q1) != qubitChoice.end()) {
+                reducedMap.emplace(q0, q1);
+            }
+        }
+    }
+}
+
+double Architecture::getAverageArchitectureFidelity(const CouplingMap& cm, const std::set<unsigned short>& qubitChoice, const Properties& props) {
+    if (props.empty()) {
+        return 0.0;
+    }
+    double result = 1.0;
+    for (const auto& [control, target]: cm) {
+        if (props.twoQubitErrorRateAvailable(control, target)) {
+            result *= (1.0 - props.getTwoQubitErrorRate(control, target));
+        }
+    }
+
+    for (const auto& qubit: qubitChoice) {
+        if (props.singleQubitErrorRate.available(qubit)) {
+            result *= (1.0 - props.getAverageSingleQubitErrorRate(qubit));
+        }
+    }
+    return result;
+}
+
+std::vector<unsigned short> Architecture::getQubitList(const CouplingMap& couplingMap) {
+    std::set<unsigned short> result{};
+    for (const auto& edge: couplingMap) {
+        result.emplace(edge.first);
+        result.emplace(edge.second);
+    }
+    return {result.begin(), result.end()};
+}
+
+bool Architecture::isConnected(const std::set<unsigned short>& qubitChoice, const CouplingMap& reducedCouplingMap) {
+    std::set<unsigned short> reachedQubits{};
+    reachedQubits.insert(*(qubitChoice.begin()));
+    dfs(*(qubitChoice.begin()), reachedQubits, reducedCouplingMap);
+    return (reachedQubits == qubitChoice);
+}
