@@ -1,10 +1,14 @@
 /*
- * This file is part of the MQT QMAP library which is released under the MIT license.
- * See file README.md or go to https://iic.jku.at/eda/research/ibm_qx_mapping/ for more information.
- */
+* This file is part of the MQT QMAP library which is released under the MIT license.
+* See file README.md or go to https://www.cda.cit.tum.de/research/ibm_qx_mapping/ for more information.
+*/
 
 #include "exact/ExactMapper.hpp"
 
+#include "Encodings/Encodings.hpp"
+#include "LogicBlock/LogicBlock.hpp"
+#include "LogicUtil/util_logicblock.hpp"
+
 void ExactMapper::map(const Configuration& settings) {
     results.config     = settings;
     const auto& config = results.config;
@@ -70,38 +74,24 @@ void ExactMapper::map(const Configuration& settings) {
             return;
         }
 
-        CouplingMap reducedCouplingMap = architecture.getCouplingMap();
-        for (const auto& edge: architecture.getCouplingMap()) {
-            if (!config.subgraph.count(edge.first) || !config.subgraph.count(edge.second)) {
-                reducedCouplingMap.erase(edge);
-            }
-        }
+        CouplingMap reducedCouplingMap{};
+        architecture.getReducedCouplingMap(config.subgraph, reducedCouplingMap);
 
         // check if the subgraph is connected
-        std::set<unsigned short> reachedQubits{};
-        reachedQubits.insert(*(config.subgraph.begin()));
-        dfs(*(config.subgraph.begin()), reachedQubits, reducedCouplingMap);
-        if (!(reachedQubits == config.subgraph)) {
+        if (!Architecture::isConnected(config.subgraph, reducedCouplingMap)) {
             std::cerr << "The subgraph is not connected." << std::endl;
             return;
         }
-        for (const auto& q: config.subgraph) {
-            qubitRange.emplace_back(q);
-        }
+        qubitRange = Architecture::getQubitList(reducedCouplingMap);
     } else {
-        for (unsigned short i = 0; i < architecture.getNqubits(); ++i) {
-            qubitRange.push_back(i);
-        }
+        qubitRange.clear();
+        auto qubitSet = architecture.getQubitSet();
+        qubitRange.reserve(qubitSet.size());
+        std::copy(qubitSet.begin(), qubitSet.end(), std::back_inserter(qubitRange));
     }
     std::vector<QubitChoice> allPossibleQubitChoices{};
     if (config.useSubsets) {
-        do {
-            QubitChoice qubitChoice{};
-            for (unsigned short i = 0; i < qc.getNqubits(); ++i) {
-                qubitChoice.insert(qubitRange.at(i));
-            }
-            allPossibleQubitChoices.push_back(qubitChoice);
-        } while (next_combination(qubitRange.begin(), qubitRange.begin() + qc.getNqubits(), qubitRange.end()));
+        allPossibleQubitChoices = architecture.getAllConnectedSubsets(qc.getNqubits());
     } else {
         QubitChoice allQubits(qubitRange.begin(), qubitRange.end());
         allPossibleQubitChoices.push_back(allQubits);
@@ -154,21 +144,10 @@ void ExactMapper::map(const Configuration& settings) {
             choiceResults.output.gates            = std::numeric_limits<unsigned long>::max();
 
             // 4) reduce coupling map
-            CouplingMap reducedCouplingMap = architecture.getCouplingMap();
-            for (const auto& edge: architecture.getCouplingMap()) {
-                if (!choice.count(edge.first) || !choice.count(edge.second)) {
-                    reducedCouplingMap.erase(edge);
-                }
-            }
-            if (reducedCouplingMap.empty()) {
-                break;
-            }
+            CouplingMap reducedCouplingMap = {};
+            architecture.getReducedCouplingMap(choice, reducedCouplingMap);
 
-            // 5) Check if E_k is connected. If yes, then possible subset found
-            std::set<unsigned short> reachedQubits{};
-            reachedQubits.insert(*(choice.begin()));
-            dfs(*(choice.begin()), reachedQubits, reducedCouplingMap);
-            if (!(reachedQubits == choice)) {
+            if (reducedCouplingMap.empty()) {
                 break;
             }
 
@@ -352,9 +331,19 @@ void ExactMapper::map(const Configuration& settings) {
 
 void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice, const CouplingMap& rcm, MappingResults& choiceResults, std::vector<std::vector<std::pair<unsigned short, unsigned short>>>& swaps, long unsigned int limit, unsigned int timeout) {
     const auto& config = results.config;
-
-    // Z3 context
-    context c;
+    using namespace logicbase;
+    // LogicBlock
+    bool              success = false;
+    logicutil::Params params;
+    params.addParam("timeout", timeout);
+    params.addParam("pb.compile_equality", true);
+    params.addParam("pp.wcnf", true);
+    params.addParam("maxres.hill_climb", true);
+    params.addParam("maxres.pivot_on_correction_set", false);
+    std::unique_ptr<LogicBlockOptimizer> lb = logicutil::getZ3LogicOptimizer(success, true, params);
+    if (!success) {
+        throw QMAPException("Could not initialize Z3 logic block optimizer");
+    }
 
     std::vector<unsigned short>                        pi(qubitChoice.begin(), qubitChoice.end());
     unsigned long long                                 piCount{};
@@ -382,40 +371,37 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
     //////////////////////////////////////////
     /// 	Boolean Variable Definitions	//
     //////////////////////////////////////////
-    /*
-	 auxilary variable declarations
-	*/
-    expr_vector auxvars(c);
 
     /*
-	 locical/physical qubit variables x_k_i_j
-	 k	before layer k
-	 i	physical qubit i
-	 j	logical qubit j
-	 number of variables: (|L|) * m * n
-	 */
-    std::vector<matrix> x{};
-    std::stringstream   x_name{};
+    locical/physical qubit variables x_k_i_j
+    k	before layer k
+    i	physical qubit i
+    j	logical qubit j
+    number of variables: (|L|) * m * n
+    */
+    LogicMatrix3D     x{};
+    std::stringstream x_name{};
     for (unsigned long k = 0; k < reducedLayerIndices.size(); ++k) {
         x.emplace_back();
         for (unsigned short Q: qubitChoice) {
-            x.back().emplace_back(c);
+            x.back().emplace_back();
             for (unsigned short q = 0; q < qc.getNqubits(); ++q) {
                 x_name.str("");
                 x_name << "x_" << k << '_' << Q << '_' << q;
-                x.back().back().push_back(c.bool_const(x_name.str().c_str()));
+                x.back().back().emplace_back(
+                        lb->makeVariable(x_name.str(), CType::BOOL));
             }
         }
     }
 
     /*
- permutation variables y_k_pi
- k	before layer k
- pi	arbitrary permutation of the m qubits
- number of variables: (|L|-1) * m!
- */
-    std::vector<std::vector<expr>> y{};
-    std::stringstream              y_name{};
+permutation variables y_k_pi
+k	before layer k
+pi	arbitrary permutation of the m qubits
+number of variables: (|L|-1) * m!
+*/
+    LogicMatrix       y{};
+    std::stringstream y_name{};
     for (unsigned long k = 1; k < reducedLayerIndices.size(); ++k) {
         y.emplace_back();
         piCount = 0;
@@ -423,106 +409,96 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
             if (skipped_pi.count(piCount) == 0 || !config.enableSwapLimits) {
                 y_name.str("");
                 y_name << "y_" << k << '_' << piCount;
-                y.back().push_back(c.bool_const(y_name.str().c_str()));
+                y.back().emplace_back(lb->makeVariable(y_name.str(), CType::BOOL));
             }
             ++piCount;
         } while (std::next_permutation(pi.begin(), pi.end()));
     }
 
-    // Z3 optimizer
-    optimize opt(c);
-    params   p(c);
-    p.set("timeout", timeout);
-    p.set("pb.compile_equality", true);
-    p.set("pp.wcnf", true);
-    p.set("maxres.hill_climb", true);
-    p.set("maxres.pivot_on_correction_set", false);
-    opt.set(p);
-
     //////////////////////////////////////////
     /// 	Consistency Constraints			//
     //////////////////////////////////////////
     if (config.encoding == Encoding::Naive) {
         for (unsigned long k = 0; k < reducedLayerIndices.size(); ++k) {
             for (unsigned long i = 0; i < qubitChoice.size(); ++i) {
-                expr rowConsistency = c.int_val(0);
+                auto rowConsistency = LogicTerm(0);
                 for (unsigned short j = 0; j < qc.getNqubits(); ++j) {
                     rowConsistency = rowConsistency +
-                                     ite(x[k][i][j], c.int_val(1), c.int_val(0));
+                                     LogicTerm::ite(x[k][i][j], LogicTerm(1), LogicTerm(0));
                 }
-                opt.add(rowConsistency.simplify() <= 1);
+                lb->assertFormula(rowConsistency <= LogicTerm(1));
             }
 
             for (unsigned short j = 0; j < qc.getNqubits(); ++j) {
-                expr colConsistency = c.int_val(0);
+                auto colConsistency = LogicTerm(0);
                 for (unsigned long i = 0; i < qubitChoice.size(); ++i) {
                     colConsistency = colConsistency +
-                                     ite(x[k][i][j], c.int_val(1), c.int_val(0));
+                                     LogicTerm::ite(x[k][i][j], LogicTerm(1), LogicTerm(0));
                 }
-                opt.add(colConsistency.simplify() == 1);
+                lb->assertFormula(colConsistency == LogicTerm(1));
             }
         }
     } else if (config.encoding == Encoding::Commander) {
         for (unsigned long k = 0; k < reducedLayerIndices.size(); ++k) {
             for (unsigned long i = 0; i < qubitChoice.size(); ++i) {
-                std::vector<expr> varIDs;
+                std::vector<LogicTerm> varIDs;
                 for (unsigned short j = 0; j < qc.getNqubits(); ++j) {
                     varIDs.push_back(x[k][i][j]);
                 }
                 if (config.commanderGrouping == CommanderGrouping::Fixed2) {
-                    opt.add(AtMostOneCMDR(varIDs, groupVars(varIDs, 2), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::atMostOneCmdr(encodings::groupVars(varIDs, 2), LogicTerm::noneTerm(), lb.get()));
                 } else if (config.commanderGrouping == CommanderGrouping::Fixed3) {
-                    opt.add(AtMostOneCMDR(varIDs, groupVars(varIDs, 3), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::atMostOneCmdr(encodings::groupVars(varIDs, 3), LogicTerm::noneTerm(), lb.get()));
                 } else if (config.commanderGrouping == CommanderGrouping::Logarithm) {
-                    opt.add(AtMostOneCMDR(varIDs, groupVars(varIDs, static_cast<std::size_t>(std::log(varIDs.size()))), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::atMostOneCmdr(encodings::groupVars(varIDs, static_cast<std::size_t>(std::log(varIDs.size()))), LogicTerm::noneTerm(), lb.get()));
                 } else if (config.commanderGrouping == CommanderGrouping::Halves) {
-                    opt.add(AtMostOneCMDR(varIDs, groupVars(varIDs, varIDs.size() / 2), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::atMostOneCmdr(encodings::groupVars(varIDs, varIDs.size() / 2), LogicTerm::noneTerm(), lb.get()));
                 }
             }
 
             for (unsigned short j = 0; j < qc.getNqubits(); ++j) {
-                std::vector<expr> varIDs;
+                std::vector<LogicTerm> varIDs;
                 for (unsigned long i = 0; i < qubitChoice.size(); ++i) {
                     varIDs.push_back(x[k][i][j]);
                 }
 
                 if (config.commanderGrouping == CommanderGrouping::Fixed2) {
-                    opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, 2), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, 2), LogicTerm::noneTerm(), lb.get()));
                 } else if (config.commanderGrouping == CommanderGrouping::Fixed3) {
-                    opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, 3), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, 3), LogicTerm::noneTerm(), lb.get()));
                 } else if (config.commanderGrouping == CommanderGrouping::Logarithm) {
-                    opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, static_cast<std::size_t>(std::log(varIDs.size()))), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, static_cast<std::size_t>(std::log(varIDs.size()))), LogicTerm::noneTerm(), lb.get()));
                 } else if (config.commanderGrouping == CommanderGrouping::Halves) {
-                    opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, varIDs.size() / 2), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, varIDs.size() / 2), LogicTerm::noneTerm(), lb.get()));
                 }
             }
         }
     } else if (config.encoding == Encoding::Bimander) {
         for (unsigned long k = 0; k < reducedLayerIndices.size(); ++k) {
             for (unsigned long i = 0; i < qubitChoice.size(); ++i) {
-                std::vector<expr>          vars;
+                std::vector<LogicTerm>     vars;
                 std::vector<unsigned long> varIDs;
                 for (unsigned short j = 0; j < qc.getNqubits(); ++j) {
                     vars.emplace_back(x[k][i][j]);
                     varIDs.emplace_back(j);
                 }
-                opt.add(AtMostOneBiMander(vars, varIDs, auxvars, c));
+                lb->assertFormula(encodings::atMostOneBiMander(vars, lb.get()));
             }
 
             for (unsigned short j = 0; j < qc.getNqubits(); ++j) { //There is no exactly one Bimander
-                std::vector<expr> varIDs;
+                std::vector<LogicTerm> varIDs;
                 for (unsigned long i = 0; i < qubitChoice.size(); ++i) {
                     varIDs.push_back(x[k][i][j]);
                 }
 
                 if (config.commanderGrouping == CommanderGrouping::Fixed2) {
-                    opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, 2), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, 2), LogicTerm::noneTerm(), lb.get()));
                 } else if (config.commanderGrouping == CommanderGrouping::Fixed3) {
-                    opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, 3), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, 3), LogicTerm::noneTerm(), lb.get()));
                 } else if (config.commanderGrouping == CommanderGrouping::Logarithm) {
-                    opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, static_cast<std::size_t>(std::log(varIDs.size()))), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, static_cast<std::size_t>(std::log(varIDs.size()))), LogicTerm::noneTerm(), lb.get()));
                 } else if (config.commanderGrouping == CommanderGrouping::Halves) {
-                    opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, varIDs.size() / 2), static_cast<int>(auxvars.size() - 1), auxvars, c));
+                    lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, varIDs.size() / 2), LogicTerm::noneTerm(), lb.get()));
                 }
             }
         }
@@ -532,12 +508,12 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
     ///		Coupling Constraints			//
     //////////////////////////////////////////
     for (unsigned long k = 0; k < reducedLayerIndices.size(); ++k) {
-        expr allCouplings = c.bool_val(true);
+        auto allCouplings = LogicTerm(true);
         for (const auto& gate: layers.at(reducedLayerIndices.at(k))) {
             if (gate.singleQubit())
                 continue;
 
-            expr coupling = c.bool_val(false);
+            auto coupling = LogicTerm(false);
             if (architecture.bidirectional()) {
                 for (const auto& edge: rcm) {
                     auto indexFC = x[k][physicalQubitIndex[edge.first]][gate.control];
@@ -554,9 +530,9 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
                     coupling = coupling || ((indexFC && indexST) || (indexFT && indexSC));
                 }
             }
-            allCouplings = allCouplings && coupling.simplify();
+            allCouplings = allCouplings && coupling;
         }
-        opt.add(allCouplings.simplify());
+        lb->assertFormula(allCouplings);
     }
 
     //////////////////////////////////////////
@@ -569,7 +545,7 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
         auto& j         = x[k];
         do {
             if (skipped_pi.count(piCount) == 0 || !config.enableSwapLimits) {
-                expr equal = c.bool_val(true);
+                auto equal = LogicTerm(true);
                 for (unsigned short Q: qubitChoice) {
                     for (unsigned short q = 0; q < qc.getNqubits(); ++q) {
                         auto before = i[physicalQubitIndex[Q]][q];
@@ -577,7 +553,7 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
                         equal       = equal && (before == after);
                     }
                 }
-                opt.add(implies(y[k - 1][internalPiCount], equal.simplify()).simplify());
+                lb->assertFormula(LogicTerm::implies(y[k - 1][internalPiCount], equal));
                 ++internalPiCount;
             }
             ++piCount;
@@ -587,21 +563,21 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
     // Allow only 1 y_k_pi to be true
     if (config.encoding == Encoding::Naive) {
         for (unsigned long k = 1; k < reducedLayerIndices.size(); ++k) {
-            expr onlyOne    = c.int_val(0);
+            auto onlyOne    = LogicTerm(0);
             piCount         = 0;
             internalPiCount = 0;
             do {
                 if (skipped_pi.count(piCount) == 0 || !config.enableSwapLimits) {
-                    onlyOne = onlyOne + ite(y[k - 1][internalPiCount], c.int_val(1), c.int_val(0));
+                    onlyOne = onlyOne + LogicTerm::ite(y[k - 1][internalPiCount], LogicTerm(1), LogicTerm(0));
                     ++internalPiCount;
                 }
                 ++piCount;
             } while (std::next_permutation(pi.begin(), pi.end()));
-            opt.add(onlyOne.simplify() == 1);
+            lb->assertFormula(onlyOne == LogicTerm(1));
         }
     } else {
         for (unsigned long k = 1; k < reducedLayerIndices.size(); ++k) {
-            std::vector<expr> varIDs;
+            std::vector<LogicTerm> varIDs;
             piCount         = 0;
             internalPiCount = 0;
             do {
@@ -612,13 +588,13 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
                 ++piCount;
             } while (std::next_permutation(pi.begin(), pi.end()));
             if (config.commanderGrouping == CommanderGrouping::Fixed2) {
-                opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, 2), -1, auxvars, c));
+                lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, 2), LogicTerm::noneTerm(), lb.get()));
             } else if (config.commanderGrouping == CommanderGrouping::Fixed3) {
-                opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, 3), -1, auxvars, c));
+                lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, 3), LogicTerm::noneTerm(), lb.get()));
             } else if (config.commanderGrouping == CommanderGrouping::Logarithm) {
-                opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, static_cast<std::size_t>(std::log(varIDs.size()))), -1, auxvars, c));
+                lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, static_cast<std::size_t>(std::log(varIDs.size()))), LogicTerm::noneTerm(), lb.get()));
             } else if (config.commanderGrouping == CommanderGrouping::Halves) {
-                opt.add(ExactlyOneCMDR(varIDs, groupVars(varIDs, varIDs.size() / 2), -1, auxvars, c));
+                lb->assertFormula(encodings::exactlyOneCmdr(encodings::groupVars(varIDs, varIDs.size() / 2), LogicTerm::noneTerm(), lb.get()));
             }
         }
     }
@@ -628,7 +604,8 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
     // cost for permutations
     piCount         = 0;
     internalPiCount = 0;
-    std::vector<std::set<WeightedVar>> weightedVars(reducedLayerIndices.size());
+    std::vector<std::set<encodings::WeightedVar>> weightedVars(reducedLayerIndices.size());
+    auto                                          cost = LogicTerm(0);
     do {
         if (skipped_pi.count(piCount) == 0 || !config.enableSwapLimits) {
             auto picost = architecture.minimumNumberOfSwaps(pi);
@@ -638,57 +615,52 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
                 picost *= GATES_OF_UNIDIRECTIONAL_SWAP;
             }
             for (unsigned long k = 1; k < reducedLayerIndices.size(); ++k) {
-                opt.add(!y[k - 1][internalPiCount], picost);
+                cost = cost + LogicTerm::ite(y[k - 1][internalPiCount], LogicTerm(static_cast<int>(picost)), LogicTerm(0));
                 if (config.useBDD)
-                    weightedVars[k].insert(WeightedVar(internalPiCount, static_cast<int>(picost)));
+                    weightedVars[k].emplace(encodings::WeightedVar(y[k - 1][internalPiCount], static_cast<int>(picost)));
             }
             ++internalPiCount;
         }
         ++piCount;
     } while (std::next_permutation(pi.begin(), pi.end()));
-
+    lb->minimize(cost);
     if (config.enableSwapLimits && config.useBDD) {
         for (unsigned long k = 1; k < reducedLayerIndices.size(); ++k) {
-            opt.add(BuildBDD(weightedVars[k], y[k - 1], auxvars, static_cast<int>(limit), c));
+            lb->assertFormula(BuildBDD(weightedVars[k], y[k - 1], static_cast<int>(limit), lb.get()));
         }
     }
 
     // cost for reversed directions
     if (!architecture.bidirectional()) {
+        cost = LogicTerm(0);
         for (unsigned long k = 0; k < reducedLayerIndices.size(); ++k) {
             for (const auto& gate: layers.at(reducedLayerIndices.at(k))) {
                 if (gate.singleQubit())
                     continue;
 
-                expr reverse = c.bool_val(true);
+                auto reverse = LogicTerm(true);
                 for (const auto& edge: rcm) {
                     auto indexFT = x[k][physicalQubitIndex[edge.first]][gate.target];
                     auto indexSC = x[k][physicalQubitIndex[edge.second]][gate.control];
                     reverse      = reverse && (!indexFT || !indexSC);
                 }
-                opt.add(reverse.simplify(), GATES_OF_DIRECTION_REVERSE);
+                cost = cost + LogicTerm::ite(reverse, LogicTerm(GATES_OF_DIRECTION_REVERSE), LogicTerm(0));
             }
         }
+        lb->minimize(cost);
     }
 
     if (config.includeWCNF) {
-        std::stringstream ss{};
-        ss << opt;
-        try {
-            c.check_error();
-        } catch (const z3::exception& e) {
-            std::cerr << "Z3 reported an exception while trying to gather the WCNF formula: " << e.msg() << std::endl;
-            std::cerr << "Most likely, this is due to the usage of Z3's atMostOne and exactlyOne constraints." << std::endl;
-            std::cerr << "This can be circumvented by using QMAP's `commander` or `bimander` encoding." << std::endl;
-        }
-        choiceResults.wcnf = ss.str();
+        choiceResults.wcnf = lb->dumpInternalSolver();
     }
 
     //////////////////////////////////////////
     /// 	Solving							//
     //////////////////////////////////////////
-    if (sat == opt.check()) {
-        model m               = opt.get_model();
+    lb->produceInstance();
+    const auto res = lb->solve();
+    if (Result::SAT == res) {
+        const auto m          = lb->getModel();
         choiceResults.timeout = results.timeout = false;
 
         // quickly determine cost
@@ -697,7 +669,6 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
         choiceResults.output.gates            = choiceResults.output.singleQubitGates + choiceResults.output.cnots;
         assert(choiceResults.output.swaps == 0U);
         assert(choiceResults.output.directionReverse == 0U);
-
         // swaps
         for (unsigned long k = 1; k < reducedLayerIndices.size(); ++k) {
             auto& i = x[k - 1];
@@ -705,17 +676,18 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
 
             for (unsigned short Q: qubitChoice) {
                 for (unsigned short q = 0; q < qc.getNqubits(); ++q) {
-                    if (eq(m.eval(i[physicalQubitIndex[Q]][q]), c.bool_val(true))) {
+                    if (m->getBoolValue(i[physicalQubitIndex[Q]][q], lb.get())) {
                         // logical qubit q was mapped to physical qubit Q
                         for (unsigned short P: qubitChoice) {
                             // and has been assigned to physical qubit P going forward
-                            if (eq(m.eval(j[physicalQubitIndex[P]][q]), c.bool_val(true))) {
+                            if (m->getBoolValue(j[physicalQubitIndex[P]][q], lb.get())) {
                                 pi[physicalQubitIndex[Q]] = P;
                             }
                         }
                     }
                 }
             }
+
             architecture.minimumNumberOfSwaps(pi, swaps.at(k));
             choiceResults.output.swaps += swaps.at(k).size();
             if (architecture.bidirectional()) {
@@ -734,7 +706,7 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
                     for (const auto& edge: rcm) {
                         auto indexFT = x[k][physicalQubitIndex[edge.first]][gate.target];
                         auto indexSC = x[k][physicalQubitIndex[edge.second]][gate.control];
-                        if (eq(m.eval(indexFT && indexSC), c.bool_val(true))) {
+                        if (m->getBoolValue(indexFT, lb.get()) && m->getBoolValue(indexSC, lb.get())) {
                             choiceResults.output.directionReverse++;
                             choiceResults.output.gates += GATES_OF_DIRECTION_REVERSE;
                         }
@@ -746,7 +718,7 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
         // save initial layout for later
         for (const auto& Q: qubitChoice) {
             for (unsigned short q = 0; q < qc.getNqubits(); ++q) {
-                bool set = eq(m.eval(x[0][physicalQubitIndex[Q]][q]), c.bool_val(true));
+                bool set = m->getBoolValue(x[0][physicalQubitIndex[Q]][q], lb.get());
                 if (set) {
                     swaps.at(0).emplace_back(std::pair<unsigned short, unsigned short>{Q, q});
                 }
@@ -756,4 +728,5 @@ void ExactMapper::coreMappingRoutine(const std::set<unsigned short>& qubitChoice
     } else {
         results.timeout = true;
     }
+    lb->reset();
 }
