@@ -1,6 +1,6 @@
 /*
  * This file is part of the MQT QMAP library which is released under the MIT license.
- * See file README.md or go to https://iic.jku.at/eda/research/ibm_qx_mapping/ for more information.
+ * See file README.md or go to https://www.cda.cit.tum.de/research/ibm_qx_mapping/ for more information.
  */
 
 #include "heuristic/HeuristicMapper.hpp"
@@ -111,6 +111,18 @@ void HeuristicMapper::map(const Configuration& ms) {
         }
     }
 
+    // infer output permutation from qubit locations
+    qcMapped.outputPermutation.clear();
+    std::size_t count = 0U;
+    for (std::size_t i = 0U; i < architecture.getNqubits(); ++i) {
+        if (qubits[i] != -1) {
+            qcMapped.outputPermutation[static_cast<dd::Qubit>(i)] = static_cast<dd::Qubit>(qubits[i]);
+        } else {
+            qcMapped.setLogicalQubitGarbage(qc.getNqubits() + count);
+            ++count;
+        }
+    }
+
     // fix single qubit gates
     if (!gatesToAdjust.empty()) {
         gateidx--; // index of last operation
@@ -153,18 +165,6 @@ void HeuristicMapper::map(const Configuration& ms) {
         }
     }
 
-    // infer output permutation from qubit locations
-    qcMapped.outputPermutation.clear();
-    std::size_t count = 0U;
-    for (std::size_t i = 0U; i < architecture.getNqubits(); ++i) {
-        if (qubits[i] != -1) {
-            qcMapped.outputPermutation[static_cast<dd::Qubit>(i)] = static_cast<dd::Qubit>(qubits[i]);
-        } else {
-            qcMapped.setLogicalQubitGarbage(qc.getNqubits() + count);
-            ++count;
-        }
-    }
-
     postMappingOptimizations(config);
     countGates(qcMapped, results.output);
     finalizeMappedCircuit();
@@ -175,6 +175,43 @@ void HeuristicMapper::map(const Configuration& ms) {
     results.timeout                    = false;
 }
 
+void HeuristicMapper::staticInitialMapping() {
+    for (const auto& gate: layers.at(0U)) {
+        if (gate.singleQubit()) {
+            continue;
+        }
+
+        for (const auto& [q0, q1]: architecture.getCouplingMap()) {
+            if (qubits.at(q0) == DEFAULT_POSITION && qubits.at(q1) == DEFAULT_POSITION) {
+                qubits.at(q0)                     = gate.control;
+                qubits.at(q1)                     = gate.target;
+                locations.at(gate.control)        = q0;
+                locations.at(gate.target)         = q1;
+                qcMapped.initialLayout.at(q0)     = gate.control;
+                qcMapped.initialLayout.at(q1)     = gate.target;
+                qcMapped.outputPermutation.at(q0) = gate.control;
+                qcMapped.outputPermutation.at(q1) = gate.target;
+                break;
+            }
+        }
+    }
+
+    // assign remaining logical qubits
+    for (unsigned short i = 0U; i < architecture.getNqubits(); ++i) {
+        if (qc.initialLayout.count(i) && locations.at(i) == DEFAULT_POSITION) {
+            for (unsigned short j = 0U; j < architecture.getNqubits(); ++j) {
+                if (qubits.at(j) == DEFAULT_POSITION) {
+                    locations.at(i)                  = j;
+                    qubits.at(j)                     = i;
+                    qcMapped.initialLayout.at(j)     = i;
+                    qcMapped.outputPermutation.at(j) = i;
+                    break;
+                }
+            }
+        }
+    }
+}
+
 void HeuristicMapper::createInitialMapping() {
     auto& config = results.config;
 
@@ -223,42 +260,7 @@ void HeuristicMapper::createInitialMapping() {
             }
             break;
         case InitialLayout::Static:
-            for (const auto& gate: layers.at(0)) {
-                if (gate.singleQubit())
-                    continue;
-
-                for (const auto& edge: architecture.getCouplingMap()) {
-                    if (qubits.at(edge.first) == DEFAULT_POSITION && qubits.at(edge.second) == DEFAULT_POSITION) {
-                        qubits.at(edge.first)                      = gate.control;
-                        qubits.at(edge.second)                     = gate.target;
-                        locations.at(gate.control)                 = edge.first;
-                        locations.at(gate.target)                  = edge.second;
-                        qcMapped.initialLayout.at(edge.first)      = gate.control;
-                        qcMapped.initialLayout.at(edge.second)     = gate.target;
-                        qcMapped.outputPermutation.at(edge.first)  = gate.control;
-                        qcMapped.outputPermutation.at(edge.second) = gate.target;
-                        break;
-                    }
-                }
-            }
-
-            // assign remaining logical qubits
-            for (unsigned short i = 0; i < architecture.getNqubits(); ++i) {
-                if (qc.initialLayout.count(i)) {
-                    if (locations.at(i) == DEFAULT_POSITION) {
-                        for (unsigned short j = 0; j < architecture.getNqubits(); ++j) {
-                            if (qubits.at(j) == DEFAULT_POSITION) {
-                                locations.at(i)                  = j;
-                                qubits.at(j)                     = i;
-                                qcMapped.initialLayout.at(j)     = i;
-                                qcMapped.outputPermutation.at(j) = i;
-                                break;
-                            }
-                        }
-                    }
-                }
-            }
-
+            staticInitialMapping();
             break;
         case InitialLayout::Dynamic:
         case InitialLayout::None:
@@ -282,6 +284,7 @@ void HeuristicMapper::mapUnmappedGates(long layer, HeuristicMapper::Node& node,
 
         if (controlLocation == DEFAULT_POSITION && targetLocation == DEFAULT_POSITION) {
             std::set<Edge> possibleEdges{};
+            // gather all edges in the architecture for which both qubits are unmapped
             for (const auto& edge: architecture.getCouplingMap()) {
                 if (qubits.at(edge.first) == DEFAULT_POSITION && qubits.at(edge.second) == DEFAULT_POSITION) {
                     possibleEdges.emplace(edge);
@@ -290,6 +293,7 @@ void HeuristicMapper::mapUnmappedGates(long layer, HeuristicMapper::Node& node,
             std::pair<unsigned short, unsigned short> chosenEdge;
 
             if (possibleEdges.empty()) {
+                // map to 2 qubits with minimal distance
                 double bestScore = std::numeric_limits<int>::max();
 
                 for (int i = 0; i < architecture.getNqubits(); i++) {
@@ -320,9 +324,12 @@ void HeuristicMapper::mapUnmappedGates(long layer, HeuristicMapper::Node& node,
         } else if (targetLocation == DEFAULT_POSITION) {
             mapToMinDistance(gate.control, gate.target);
         }
+        // TODO: use HeuristicMapper::Node::updateHeuristicCost instead to have heuristic function implementation isolated to one place
         node.costHeur = std::max(node.costHeur, distanceOnArchitectureOfLogicalQubits(gate.control, gate.target));
     }
 
+    // if all considered qubit pairs are mapped next to each other, the mapping is done
+    // (at most separated by 1 edge possibly directed in the wrong direction requiring 1 reversing gate)
     node.done = node.costHeur <= COST_DIRECTION_REVERSE;
 }
 
@@ -365,6 +372,7 @@ HeuristicMapper::Node HeuristicMapper::AstarMap(long layer) {
     Node result = nodes.top();
     nodes.pop();
 
+    // clear nodes
     while (!nodes.empty()) {
         nodes.pop();
     }
@@ -379,6 +387,7 @@ void HeuristicMapper::expandNode(const std::vector<unsigned short>& consideredQu
         used_swaps.emplace_back(architecture.getNqubits());
     }
 
+    // set up new teleportation qubits
     std::set<Edge> perms = architecture.getCouplingMap();
     architecture.getCurrentTeleportations().clear();
     architecture.getTeleportationQubits().clear();
